# Copyright 2024 Zoé Cassiopée Gauthier.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

import math
from enum import Flag, auto
from typing import final

from rich.progress import Progress

from pogo_scaled_estimators.pokebattler_proxy import MovesetResult, PokebattlerProxy, Raid
from pogo_scaled_estimators.utilities import format_move_name, format_pokemon_name


class Filter(Flag):
    NO_FILTER = 0
    DISALLOW_LEGACY_MOVES = auto()
    DISALLOW_MEGA_POKEMON = auto()
    DISALLOW_SHADOW_POKEMON = auto()
    DISALLOW_LEGENDARY_POKEMON = auto()


@final
class Calculator:
    def __init__(self, attacker_types: list[str], filters: Filter = Filter.NO_FILTER) -> None:
        self.attacker_types = attacker_types
        self.filters = filters
        self._pokebattler_proxy = PokebattlerProxy()
        self._progress = Progress()

    def calculate(self, level: int = 40, party: int = 1) -> None:
        raid_bosses = self._pokebattler_proxy.raid_bosses(self.attacker_types)
        attackers = {
            attacker: {"RAID_LEVEL_3": [], "RAID_LEVEL_5": [], "RAID_LEVEL_MEGA": []}
            for attacker in self._pokebattler_proxy.with_charged_moves(self.attacker_types)
            if self._allowed_attacker(attacker)
        }

        with self._progress:
            total_defenders = sum(len(defenders) for defenders in raid_bosses.values())
            task = self._progress.add_task("Simulating raids...", total=total_defenders)

            for raid_tier, defenders in raid_bosses.items():
                if raid_tier == "RAID_LEVEL_MEGA_5":
                    simplified_raid_tier = "RAID_LEVEL_MEGA"
                elif raid_tier == "RAID_LEVEL_ULTRA_BEAST":
                    simplified_raid_tier = "RAID_LEVEL_5"
                else:
                    simplified_raid_tier = raid_tier
                for defender in defenders:
                    defender_type = self._pokebattler_proxy.pokemon_type(defender)
                    self._progress.console.log(f"vs {format_pokemon_name(defender, defender_type)}...")
                    self._progress.update(task, advance=1)
                    raid = Raid(raid_tier, defender, level, party)
                    results = {
                        attacker: [moveset for moveset in movesets if self._allowed_moveset(attacker, moveset)]
                        for attacker, movesets in self._pokebattler_proxy.simulate(raid).items()
                        if attacker in attackers
                    }
                    if not results:
                        continue
                    best_movesets = {
                        attacker: min(movesets, key=lambda moveset: moveset.estimator)
                        for attacker, movesets in results.items()
                        if movesets
                    }
                    best_estimator = min(best_movesets.values(), key=lambda moveset: moveset.estimator).estimator
                    for attacker, moveset in best_movesets.items():
                        attackers[attacker][simplified_raid_tier].append(moveset.scale(best_estimator))

        ase_by_attacker: list[tuple[str, str, str, float]] = []
        for attacker, raid_tier_results in attackers.items():
            fast_move, charged_move = self._best_moves(raid_tier_results)
            ase = self._ase(raid_tier_results)
            if not math.isfinite(ase):
                continue
            ase_by_attacker.append((attacker, fast_move, charged_move, self._ase(raid_tier_results)))
        ase_by_attacker.sort(key=lambda item: item[3])
        for attacker, fast_move, charged_move, ase in ase_by_attacker:
            attacker_type = self._pokebattler_proxy.pokemon_type(attacker)
            fast_move_type = self._pokebattler_proxy.find_move(fast_move).typing
            charged_move_type = self._pokebattler_proxy.find_move(charged_move).typing
            self._progress.console.print(
                f"[bold]{format_pokemon_name(attacker, attacker_type)}[/bold] ({format_move_name(fast_move, fast_move_type)}/{format_move_name(charged_move, charged_move_type)}): {ase:.2f}"
            )

    def _allowed_attacker(self, pokemon_id: str) -> bool:
        if Filter.DISALLOW_MEGA_POKEMON in self.filters and ("_MEGA" in pokemon_id or "_PRIMAL" in pokemon_id):
            return False
        if Filter.DISALLOW_SHADOW_POKEMON in self.filters and "_SHADOW_FORM" in pokemon_id:
            return False
        pokemon = self._pokebattler_proxy.find_pokemon(pokemon_id)
        if Filter.DISALLOW_LEGENDARY_POKEMON in self.filters and pokemon.get("rarity") in [
            "POKEMON_RARITY_LEGENDARY",
            "POKEMON_RARITY_MYTHIC",
            "POKEMON_RARITY_ULTRA_BEAST",
        ]:
            return False
        return True

    def _allowed_moveset(self, pokemon_id: str, moveset: MovesetResult) -> bool:
        charged_move = self._pokebattler_proxy.find_move(moveset.charged_move)
        if charged_move.typing not in self.attacker_types:
            return False
        if Filter.DISALLOW_LEGACY_MOVES in self.filters:
            pokemon = self._pokebattler_proxy.find_pokemon(pokemon_id)
            if moveset.fast_move in pokemon.get("eliteQuickMove", []):
                return False
            charged_move_name = moveset.charged_move.removesuffix("_PLUS_PLUS")
            if charged_move_name in pokemon.get("eliteCinematicMove", []):
                return False
        return True

    def _best_moves(self, moveset_results_by_tier: dict[str, list[MovesetResult]]) -> tuple[str, str]:
        movesets: set[tuple[str, str]] = set()
        for moveset_results in moveset_results_by_tier.values():
            movesets.update((moveset.fast_move, moveset.charged_move) for moveset in moveset_results)
        if not movesets:
            return ("MOVE_NONE", "MOVE_NONE")
        return min(movesets, key=lambda moveset: self._ase(moveset_results_by_tier, only=moveset))

    def _ase(self, moveset_results_by_tier: dict[str, list[MovesetResult]], only=None) -> float:
        try:
            return (
                0.15 * self._average_estimator(moveset_results_by_tier["RAID_LEVEL_3"], only)
                + 0.50 * self._average_estimator(moveset_results_by_tier["RAID_LEVEL_5"], only)
                + 0.35 * self._average_estimator(moveset_results_by_tier["RAID_LEVEL_MEGA"], only)
            )
        except ZeroDivisionError:
            return float("inf")

    def _average_estimator(self, moveset_results: list[MovesetResult], only: tuple[str, str] | None = None) -> float:
        if only:
            moveset_results = [m for m in moveset_results if m.fast_move == only[0] and m.charged_move == only[1]]
        return sum(moveset.estimator for moveset in moveset_results) / len(moveset_results)
